/*

This file is part of aiorOS

aiorOS - 32bit Operating System.
Copyright (C) 2005  Salvatore Iengo <salvatoreiengo@hotmail.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include <task.h>
#include <fs.h>
#include <scheduler.h>
#include <swapper.h>
#include <std/string.h>

union task_t *tasksTable[TASKS_MAX_NUM];
static union task_t *taskHead = NULL;
int tasksCount = 0;

static struct TSS_t makeTSS(dword code, dword pageDir, word LDTentry, dword ustack, dword kstack);
static void initTasksTable(void);
static int getFreePid(void);
static int linkTask(union task_t *task);
static void unlinkTask(union task_t *task);
static dword push_args_to_stack_page(char *base, int argc, char *argv[]);
static union task_t *create_task_struct(void *buffer, void *pdbr, dword esp);
static void destroy_task_struct(union task_t *task);


void initTasks(void){
   initTasksTable();
}

int tsk_create_process(char *path, int argc, char *argv[]){
   char name[64];
   union task_t *task;
   PageDirEntry *pd;
   PageTableEntry *pt, *pts;
   void *ps;
   dword info, esp;
   int pid, size;

   if(!fs_exists(path, &size))
      return -1;

   CLI();

   strcpy(name, path);

   /* kms page */
   task = getPage();
   /* page dir */
   pd = getPage();
   /* page dir[0] 
     page table 
  */
   pt = getPage();
   /* stack page
     table
  */
   pts = getPage();
   /* stack page*/
   ps = getPage();

   if(!task || !pd || !pt || !pts || !ps)
      swapper(5);/*invoke swapper requesting 5 pages*/

   esp=USER_MODE_STACK;
   if(storePageDirRegister()!=KERNEL_CACHE){
   /* use old page dir as kernel buffer */
      esp = push_args_to_stack_page(KERNEL_CACHE, argc, argv);
   }

   info = BEGIN_PHYS_ADDR(0);

   pt = create_page_table(pt, PAGE_TYPE_KERNEL, PAGE_MODE_USER);
   pd = create_page_dir(pd, pt);
   task = create_task_struct(task, pd, esp);
   pts = create_page_table(pts, PAGE_TYPE_EMPTY, PAGE_MODE_USER);
   /* user stack */
   pd[1022] = PAGE_DIR_ENTRY(PAGE_ENTRY_PRESENT,PAGE_MODE_USER,((dword)pts)>>12);
   pts[1023] = PAGE_TABLE_ENTRY(PAGE_ENTRY_PRESENT,PAGE_MODE_USER,((dword)ps)>>12);

   if(esp!=USER_MODE_STACK){
      memcpy(ps, KERNEL_CACHE, PAGE_SIZE);
   }

   strcpy(task->Info.name, name);
   task->Info.Address.size = size;

   share_task_struct((pid=task->Info.pid), task);

   task = task->Info.Address.paged;
   END_PHYS_ADDR(info, 0);

   linkTask(task);

   STI();

   return pid;
}

static struct TSS_t makeTSS(dword code, dword pageDir, word LDTentry, dword ustack, dword kstack)
{
   struct TSS_t TSS;
   int i;
   for(i=0;i<sizeof(struct TSS_t);i++)
      ((byte*)(&TSS))[i]=0;

   TSS.CS = SEG_SEL(0,TI_LDT,GDT_DPL3);  /*01 | 1 11*/
   TSS.DS = SEG_SEL(1,TI_LDT,GDT_DPL3); /*10 | 1 11*/
   TSS.SS = SEG_SEL(1,TI_LDT,GDT_DPL3);
   TSS.ES = SEG_SEL(1,TI_LDT,GDT_DPL3);
   TSS.FS = SEG_SEL(1,TI_LDT,GDT_DPL3);
   TSS.GS = SEG_SEL(1,TI_LDT,GDT_DPL3);
   TSS.ESP = ustack;
   TSS.LDT = SEG_SEL(LDTentry,TI_GDT,GDT_DPL3);
   TSS.EFLAGS = 0x200; /*IOPL=0 + IF=1*/
   TSS.EIP = (dword)(code);
   TSS.CR3 = pageDir;
   TSS.SS0 = SEG_SEL(2,TI_GDT,GDT_DPL0);
   TSS.ESP0 = kstack;
   TSS.IOMAP = 0xDFFF; /*IOMAP (see intel manual)*/

  return TSS;
}

static void initTasksTable(void){
   int i;
   for(i=0;i<TASKS_MAX_NUM;i++)
      tasksTable[i]=NULL;
   tasksCount = 0;
}

static int start = -1;
static int getFreePid(void){
   int i;
   int end;

   start = (start+1)%TASKS_MAX_NUM;
   end = (start>0)? start - 1 : TASKS_MAX_NUM - 1;

   /*find free pid from hash table*/
   for(i=start;i!=end;i=(i+1)%TASKS_MAX_NUM){
      if(tasksTable[i]==NULL){
         /* paged address */
         tasksTable[i] = (union task_t *)TASK_STRUCT_ADDRESS(i);
         return i;
      }
   }
  return -1;
}

/* treat paged address only */
static int linkTask(union task_t *task){
   task->Info.Link.next = task;
   task->Info.Link.prev = task;

   /*list insert*/
   if(!taskHead)
      taskHead = task;
   else{
      task->Info.Link.next = taskHead->Info.Link.next;
      task->Info.Link.prev = taskHead;
      taskHead->Info.Link.next->Info.Link.prev = task;
      taskHead->Info.Link.next = task;
   }
   return task->Info.pid;
}

static void unlinkTask(union task_t *task){
   if(task==NULL)
      return;

   if(tasksTable[task->Info.pid]!=NULL){
      tasksTable[task->Info.pid]=NULL; tasksCount--;}

   task->Info.Link.prev->Info.Link.next = task->Info.Link.next;
   task->Info.Link.next->Info.Link.prev = task->Info.Link.prev;

   if(task == taskHead)
      taskHead = task->Info.Link.next;
}

/*
   Returns esp position (relative
   to stack page) after arguments
   pushing.
*/
static dword push_args_to_stack_page(char *page, int argc, char *argv[])
{
   int i,j, len;
   char *esp = (char*)page+PAGE_SIZE-1;
   char *esp_limit = esp+1;

   if(argc == 0 || argv == NULL){
      int *iesp = (int*)esp;
      *--iesp=0; /* argv */
      *--iesp=0; /* argc */
      iesp--;
     return (dword)(((char*)USER_MODE_STACK) - (esp_limit - (char*)iesp - 1));
   }

   /* copying argv strings */
   for(i=argc-1;i>=0;i--){
      len = strlen(argv[i]);
      *esp--=0; /*C terminating string char*/
      for(j=len-1;j>=0;j--){
         *esp--=argv[i][j];
      }
   }
   esp++;

   /* align to dword boundaries */
   if((esp_limit - esp) % 4)
      esp -= (4-((esp_limit - esp) % 4));

   /* setting up string pointers */
   for(i=argc-1,len=0;i>=0;i--){
      *((int*)esp-1) =(int)(((char*)USER_MODE_STACK) - (strlen(argv[i]) + len));
      esp-=sizeof(int);
      len += (strlen(argv[i]))+1;
   }
   /* write argv and argc */
   *((int*)esp-1) = (int)(((char*)USER_MODE_STACK) - (esp_limit - esp - 1));
   *((int*)esp-2) = argc;
   esp-=(int)(sizeof(int)*3);

   return (dword)(((char*)USER_MODE_STACK) - (esp_limit - esp - 1));
}

static union task_t *create_task_struct(void *buffer, void *pdbr, dword esp){
   int pid, tss;
   union task_t *task = (union task_t*)buffer;

   if((pid = getFreePid())<0)
      goto error;

   task->Info.pid = pid;

   task->Info.System.TSS = makeTSS(
        (dword)0,
        (dword)pdbr,
        GLOBAL_LDT,
        esp, /*user stack must be scaled*/
        KERNEL_MODE_STACK(pid) /*kstack not to be scaled*/
        );

   tss = setGDTgate(
        getFreeGDTgate(),
        makeGDTgate(
             sizeof(task->Info.System.TSS)-1,
             TASK_STRUCT_ADDRESS(pid),
             GDT_G_1B,GDT_TSS,
             GDT_DPL3
             )
        );

   if(tss<0)
      goto error;

   task->Info.Address.paged = (union task_t*)TASK_STRUCT_ADDRESS(pid);

   task->Info.System.tssIdx = tss;
   task->Info.System.ldtIdx = GLOBAL_LDT;
   task->Info.status = RUNNING;
   task->Info.Priority.dynamic = QUANTUM;
   task->Info.Priority.nice = 0;
   task->Info.System.tssIdx = tss;
   task->Info.System.ldtIdx = GLOBAL_LDT;
   task->Info.Address.base = (unsigned char*)KERNEL_MEM_LIMIT;
   task->Info.Address.physical = buffer;

  /* returning task struct
    paged address
  */

  return task;
error:
  return NULL;
}

void destroyTask(int pid){
   destroy_task_struct(pid2task(pid));
}

static void destroy_task_struct(union task_t *task){
   int i,j;
   dword info;
   PageDirEntry *dir = (PageDirEntry*)storePageDirRegister();
   PageTableEntry *table;

   unlinkTask(task);
   freeGDTgate(task->Info.System.tssIdx);

   task = (union task_t *)task->Info.Address.physical;
   info = BEGIN_PHYS_ADDR(1);

   for(i=0;i<1023;i++){
      if(dir[i].Present){
         table = (PageTableEntry*)(((dword)(dir[i].PageTableBaseAddress))<<12);
         for(j=(i==0)?KERNEL_PAGE_SIZE:0;j<1024;j++){
            if(table[j].Present){
               freePage((void*)(((dword)(table[j].PageBaseAddress))<<12));
            }
         }
         freePage(table);
      }
   }
   freePage(dir);
   freePage(task);

   END_PHYS_ADDR(info,1);
}
